# -*- coding: utf-8 -*-
import law
import luigi
import logging
import os
import re

from functools import partial
from itertools import product
from subprocess import PIPE

from law.util import colored

from analysis.tasks.Combine import Combine

from analysis.framework import Task, SpaceSeparatedListParameter
from analysis.util import interruptable_popen, working_directory

logger = logging.getLogger(__name__)
working_directory = partial(working_directory, logger=logger)


class Recombine(Task, law.LocalWorkflow):
    merge_dir = luigi.Parameter(default='grids')

    default_channels = ["LO", "R", "V", "RRa", "RRb", "RV", "VV", "NLO", "NLO_only", "NNLO", "NNLO_only"]
    channels = SpaceSeparatedListParameter(
        default=default_channels,
        cls=luigi.Parameter,
        choices=("LO", "R", "V", "RRa", "RRb", "RV", "VV", "NLO", "NLO_only", "NNLO", "NNLO_only")
    )
    observables = luigi.ListParameter()

    cores = luigi.Parameter()
    combine_ini = luigi.Parameter()

    def create_branch_map(self):
        """Branch data dicts. Each branch should correspond to exactly one output."""
        return {
            i : {'channel': channel, 'observable': obs}
            for i, (channel, obs) in enumerate(product(
                self.channels,
                self.observables
            ))
        }

    def workflow_requires(self):
        return {'combine': Combine()}

    def requires(self):
        return Combine()

    def output(self):
        out_dir = os.getenv("ANALYSIS_PATH") or os.getcwd()
        return law.LocalFileTarget(
            '{}/{}/{}/Combined/Final/{}.{}.{}.dat'.format(
                out_dir,
                self.merge_dir, self.name, self.process, self.branch_data['channel'],
                self.branch_data['observable']
            )
        )

    def run(self):
        # backported `enum34` is incompatible with Python > 3.4
        # -> remove it from PYTHONPATH
        env = os.environ.copy()
        env["PYTHONPATH"] = os.pathsep.join(p for p in env["PYTHONPATH"].split(os.pathsep) if not re.search('enum34', p))

        # ensure output parent directory exists
        logger.debug("Creating output parent directory: %s", self.output().parent.path)
        self.output().parent.touch()

        outfile = self.output().path
        errfile = re.sub('.dat$', '.err', outfile)

        rundirpath = os.getcwd()
        combine_ini = os.path.abspath(self.combine_ini)

        mergedir = '{}/{}'.format(self.merge_dir, self.name)

        with working_directory(mergedir):

            # DS: is this an output file? If not, who produces it?
            # KR: This is an output file of the Combine task.
            #     The product of this task are .dat files with the final statistical uncertainties
            #     after merging that are read in by AddStatunc and written into the v25 grids.
            #     They are not needed for the merging steps, where only the APPLfast.txt weight files are required.
            weightfile = 'Combined/Final/{}.{}.APPLfast.txt'.format(
                self.branch_data['channel'], self.branch_data['observable']
            )

            command_line_args = [
                'nnlojet-combine.py', '-C', combine_ini, '--APPLfast', weightfile
            ]
            logger.info("Running command: %s",
                colored(' '.join(command_line_args), 'yellow'))
            code, out, error = interruptable_popen(command_line_args,
                                                   stdout=PIPE,
                                                   stderr=PIPE,
                                                   env=env)

            logger.debug("Return code: %d", code)
            logger.debug("Standard output:\n%s\n", out)
            logger.debug("Standard error:\n%s\n", error)

        if error:
            logger.debug("Dumping standard error to file:  %s", errfile)
            with open(errfile, 'w') as f:
                f.write(error)

        if (code == 0):
            logger.info("nnlojet-combine status: SUCCESS")
            logger.debug("Dumping standard out to file:  %s", outfile)
            self.output().dump(out, formatter="text")
        else:
            error_message = "nnlojet-combine status: FAIL (exit code {})".format(code)
            logger.error(error_message)
            raise RuntimeError(error_message)
